"""Log configuration for Notebooks."""

__all__ = ["IPythonHandler", "forward_lsst_log"]

import html
import logging

from IPython.display import HTML, display

try:
    import lsst.log as lsstLog
except ImportError:
    lsstLog = None

# Each log level will have the level name use a different color to
# ensure that warning log messages can be seen more easily.
_level_colors = {
    "CRITICAL": "var(--jp-error-color0)",
    "ERROR": "var(--jp-error-color2)",
    "WARNING": "var(--jp-warn-color0)",
    "INFO": "var(--jp-info-color0)",
    "VERBOSE": "var(--jp-info-color3)",
    "DEBUG": "var(--jp-success-color0)",
    "DEFAULT": "var(--jp-success-color3)",
}

# CSS style for the PRE block to use for displaying log messages.
# The margin prevents excessive space between log messages.
# The border gives a vertical bar at the start of a log message to improve
# delineation between log output and standard output.
# The left padding gives space between that bar and the first character.
# By default PRE blocks do not use a font size that matches the print()
# output so force it here.
_pre_style = """margin: 0.1em;
padding-left: 0.25em;
border-left-style: solid;
font-family: var(--jp-code-font-family);
font-size: var(--jp-code-font-size);
line-height: var(--jp-code-line-height);
"""


class IPythonHandler(logging.Handler):
    """Special log handler for IPython Notebooks.
    This log handler emits log messages as formatted HTML output with
    the following content:
    * The name of the logger.
    * The level of the log message, color coded based on severity.
    * The message itself.
    It can be enabled (forcing other log handlers to be removed) with:
    .. code-block:: python
       logging.basicConfig(level=logging.INFO, force=True,
                           handlers=[IPythonHandler()])
    """

    def emit(self, record: logging.LogRecord) -> None:
        name_color = "var(--jp-warn-color2)"
        level_color = _level_colors.get(
            record.levelname, _level_colors["DEFAULT"]
        )
        message = html.escape(record.getMessage())
        name_msg = f'<span style="color: {name_color}">{record.name}</span>'
        level_msg = (
            f'<span style="color: {level_color}">{record.levelname}</span>'
        )
        text = (
            f'<pre style="{_pre_style}">{name_msg} {level_msg}: '
            + f"{message}</pre>"
        )
        display(HTML(text))


def forward_lsst_log(level: str):
    """Forward ``lsst.log`` level messages to Python logging.
    Parameters
    ----------
    level : `str`
        The level name to forward.
    """
    if lsstLog is not None:
        lsstLog.configure_pylog_MDC(level, MDC_class=None)
        lsstLog.usePythonLogging()



"""Utility functions for LSST JupyterLab notebook environment
"""
import os
import urllib
from pathlib import Path
from typing import Optional

import bokeh.io
from kubernetes import client, config


def format_bytes(n) -> str:
    """Format bytes as text

    >>> format_bytes(1)
    '1 B'
    >>> format_bytes(1234)
    '1.23 kB'
    >>> format_bytes(12345678)
    '12.35 MB'
    >>> format_bytes(1234567890)
    '1.23 GB'
    >>> format_bytes(1234567890000)
    '1.23 TB'
    >>> format_bytes(1234567890000000)
    '1.23 PB'

    (taken from dask.distributed, where it is not exported)
    """
    if n > 1e15:
        return "%0.2f PB" % (n / 1e15)
    if n > 1e12:
        return "%0.2f TB" % (n / 1e12)
    if n > 1e9:
        return "%0.2f GB" % (n / 1e9)
    if n > 1e6:
        return "%0.2f MB" % (n / 1e6)
    if n > 1e3:
        return "%0.2f kB" % (n / 1000)
    return "%d B" % n


def get_hostname() -> str:
    """Utility function to return hostname or, failing that, "localhost"."""
    return os.environ.get("HOSTNAME") or "localhost"


def show_with_bokeh_server(obj):
    """Method to wrap bokeh with proxy URL"""

    def jupyter_proxy_url(port):
        """
        Callable to configure Bokeh's show method when a proxy must be
        configured.

        If port is None we're asking about the URL
        for the origin header.

        https://docs.bokeh.org/en/latest/docs/user_guide/jupyter.html
        """
        base_url = os.environ["EXTERNAL_INSTANCE_URL"]
        host = urllib.parse.urlparse(base_url).netloc

        # If port is None we're asking for the URL origin
        # so return the public hostname.
        if port is None:
            return host

        service_url_path = os.environ["JUPYTERHUB_SERVICE_PREFIX"]
        proxy_url_path = "proxy/%d" % port

        user_url = urllib.parse.urljoin(base_url, service_url_path)
        full_url = urllib.parse.urljoin(user_url, proxy_url_path)
        return full_url

    bokeh.io.show(obj, notebook_url=jupyter_proxy_url)


def get_pod():
    """Get pod record.  Throws an error if you're not running in a cluster."""
    config.load_incluster_config()
    api = client.CoreV1Api()
    namespace = "default"
    with open(
        "/var/run/secrets/kubernetes.io/serviceaccount/namespace", "r"
    ) as f:
        namespace = f.readlines()[0]
    pod = api.read_namespaced_pod(get_hostname(), namespace)
    return pod


def get_node() -> str:
    """Extract node name from pod."""
    return get_pod().spec.node_name


def get_digest() -> str:
    """Extract image digest from pod, if we can."""
    digest = ""
    try:
        img_id = get_pod().status.container_statuses[0].image_id
        # host/owner/repo@sha256:hash
        return (img_id.split("@")[-1]).split(":")[-1]
    except Exception:
        return ""  # We will just return the empty string
    return digest


def get_access_token(tokenfile=None, log=None) -> Optional[str]:
    """Determine the access token from the mounted configmap (nublado2),
    secret (nublado1), or environment (either).  Prefer the mounted version
    since it can be updated, while the environment variable stays at whatever
    it was when the process was started."""
    tok = None
    if tokenfile:
        # If a path was specified, trust it.
        tok = Path(tokenfile).read_text()
        tried_path = tokenfile
    else:
        # Try the default token paths, nublado2 first, then nublado1
        n2_tokenfile = "/opt/lsst/software/jupyterlab/environment/ACCESS_TOKEN"
        tried_path = n2_tokenfile
        token_path = Path(n2_tokenfile)
        try:
            tok = token_path.read_text()
        except Exception:
            # OK, it's not mounted.  Fall back to the environment.
            pass
    if not tok:
        tok = os.environ.get("ACCESS_TOKEN", None)
    if not tok:
        raise ValueError(
            f"Could not find token in env:ACCESS_TOKEN nor in {tried_path}"
        )
    return tok



import os
from typing import Optional

import pyvo
import pyvo.auth.authsession
import requests
from deprecated import deprecated

from .utils import get_access_token


def _get_tap_url() -> str:
    tapurl = os.getenv("EXTERNAL_TAP_URL")
    if not tapurl:
        tapurl = (os.getenv("EXTERNAL_INSTANCE_URL") or "") + (
            os.getenv("TAP_ROUTE") or "/api/tap"
        )
    return tapurl


def _get_auth() -> Optional[pyvo.auth.authsession.AuthSession]:
    tap_url = _get_tap_url()
    s = requests.Session()
    tok = get_access_token()
    if not tok:
        return None
    s.headers["Authorization"] = "Bearer " + tok
    auth = pyvo.auth.authsession.AuthSession()
    auth.credentials.set("lsst-token", s)
    auth.add_security_method_for_url(tap_url, "lsst-token")
    auth.add_security_method_for_url(tap_url + "/sync", "lsst-token")
    auth.add_security_method_for_url(tap_url + "/async", "lsst-token")
    auth.add_security_method_for_url(tap_url + "/tables", "lsst-token")
    return auth


def get_tap_service() -> pyvo.dal.TAPService:
    return pyvo.dal.TAPService(_get_tap_url(), _get_auth())


@deprecated(reason="Please use get_tap_service()")
def get_catalog() -> pyvo.dal.TAPService:
    return get_tap_service()


def retrieve_query(query_url) -> pyvo.dal.AsyncTAPJob:
    return pyvo.dal.AsyncTAPJob(query_url, _get_auth())



"""All of this stuff moved between JL2 and JL3.  Export these functions and
methods again, but with deprecated warnings so people know how to use them
correctly.
"""
from deprecated import deprecated

import lsst.rsp

from .utils import get_node


@deprecated(reason="Please use lsst.rsp.format_bytes()")
def format_bytes(n):
    return lsst.rsp.format_bytes(n)


@deprecated(reason="Please use lsst.rsp.get_tap_service()")
def get_catalog():
    return lsst.rsp.get_tap_service()


@deprecated(reason="Please use lsst.rsp.get_tap_service()")
def get_tap_service():
    return lsst.rsp.get_tap_service()


@deprecated(reason="Please use lsst.rsp.retrieve_query()")
def retrieve_query(query_url):
    return lsst.rsp.retrieve_query(query_url)


@deprecated(reason="Please use lsst.rsp.get_hostname()")
def get_hostname():
    return lsst.rsp.get_hostname()


@deprecated(reason="Please use lsst.rsp.show_with_bokeh_server()")
def show_with_bokeh_server(obj):
    return lsst.rsp.show_with_bokeh_server(obj)


@deprecated(reason="Please use lsst.rsp.Forwarder")
class Forwarder(lsst.rsp.Forwarder):
    def __init__(self, args, **kwargs):
        return super().__init__()


__all__ = [
    Forwarder,
    format_bytes,
    get_catalog,
    get_node,
    get_tap_service,
    retrieve_query,
    get_hostname,
    show_with_bokeh_server,
]



from deprecated import deprecated

import lsst.rsp


@deprecated(reason="Please use lsst.rsp.get_tap_service()")
def get_catalog():
    return lsst.rsp.get_tap_service()


@deprecated(reason="Please use lsst.rsp.retrieve_query()")
def retrieve_query(query_url):
    return lsst.rsp.retrieve_query(query_url)



"""
Collection of utilities, formerly in rsp_jupyter_utils.lab and
rsp_jupyter_utils.helper
"""
from .utils import (
    format_bytes,
    get_access_token,
    get_digest,
    get_hostname,
    get_node,
    get_pod,
    show_with_bokeh_server,
)
from .logging import IPythonHandler, forward_lsst_log
from .forwarder import Forwarder
from .catalog import get_catalog, get_tap_service, retrieve_query
from importlib.metadata import PackageNotFoundError, version

__version__: str
"""The application version string of (PEP 440 / SemVer compatible)."""

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"


__all__ = [
    __version__,
    Forwarder,
    IPythonHandler,
    format_bytes,
    forward_lsst_log,
    get_access_token,
    get_catalog,
    get_digest,
    get_node,
    get_pod,
    get_tap_service,
    retrieve_query,
    get_hostname,
    show_with_bokeh_server,
]



import logging
from typing import Any, List, Optional

import maproxy.iomanager
import maproxy.proxyserver
import tornado.netutil


class Forwarder(maproxy.proxyserver.ProxyServer):
    """This creates a TCP proxy server running on a randomly-assigned
    dynamic local port.  Pass it the target host and port to construct it.
    """

    _ioloop = None
    _thread = None
    _logger = None
    _bind_addresses: List[Any] = []  # Fix this when we correct typing.

    def __init__(self, *args, **kwargs):
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Creating TCP Forwarder")
        sockets = tornado.netutil.bind_sockets(0, "")
        ioloop = None
        save_ioloop = None
        if "ioloop" in kwargs:
            self._logger.debug("IOLoop specified; saving current")
            ioloop = kwargs["ioloop"]
            save_ioloop = tornado.ioloop.IOLoop.current()
            if ioloop != save_ioloop:
                self._logger.debug("Switching IOLoop")
                ioloop.make_current()
            del kwargs["ioloop"]
        super().__init__(*args, **kwargs)
        self.add_sockets(sockets)
        self.bind_addresses = [x.getsockname()[:2] for x in sockets]
        self._ioloop = tornado.ioloop.IOLoop.current()
        if save_ioloop and save_ioloop != self._ioloop:
            self._logger.debug("Restoring IOLoop")
            save_ioloop.make_current()
        self._logger.debug("TCP Forwarder created")

    def get_ports(self):
        """Returns a list of the ports the Forwarder is listening to."""
        return list(set(x[1] for x in self.bind_addresses))

    def get_port(self) -> Optional[int]:
        """Returns the first port the Forwarder is listening to, or None."""
        rval = self.get_ports()
        if rval and len(rval) > 0:
            return rval[0]
        return None



from deprecated import deprecated

import lsst.rsp


@deprecated(reason="Please use lsst.rsp.show_with_bokeh_server()")
def show_with_bokeh_server(obj):
    return lsst.rsp.show_with_bokeh_server(obj)
